@@ -236,6 +236,14 @@ def _makeReply(irc, msg, s,
     # Finally, we'll return the actual message.
     ret = msgmaker(target, s)
     ret.tag('inReplyTo', msg)
+    if 'msgid' in msg.server_tags \
+            and conf.supybot.protocols.irc.experimentalExtensions() \
+            and 'message-tags' in irc.state.capabilities_ack:
+        # In theory, msgid being in server_tags implies message-tags was
+        # negotiated, but the +reply spec requires it explicitly. Plus, there's
+        # no harm in doing this extra check, in case a plugin is replying
+        # across network (as it may happen with '@network command').
+        ret.server_tags['+draft/reply'] = msg.server_tags['msgid']
     return ret
 
 def error(*args, **kwargs):
@@ -1410,14 +1418,15 @@ class PluginMixin(BasePlugin, irclib.IrcCallback):
                 noIgnore = self.noIgnore(irc, msg)
             else:
                 noIgnore = self.noIgnore
-            if noIgnore or \
-               not ircdb.checkIgnored(msg.prefix, msg.channel) or \
-               not ircutils.isUserHostmask(msg.prefix):  # Some services impl.
+            if (noIgnore or
+               not msg.prefix or  # simulated echo message
+               not ircdb.checkIgnored(msg.prefix, msg.channel) or
+               not ircutils.isUserHostmask(msg.prefix)):  # Some services impl.
                 self.__parent.__call__(irc, msg)
         else:
             self.__parent.__call__(irc, msg)
 
-    def registryValue(self, name, channel=None, network=None, value=True):
+    def registryValue(self, name, channel=None, network=None, *, value=True):
         if isinstance(network, bool):
             # Network-unaware plugin that uses 'value' as a positional
             # argument.
