@@ -73,12 +73,12 @@ class FunctionsTestCase(SupyTestCase):
         self.assertNotEqual(f(irc, msg, '$today'), '$today')
         self.assertNotEqual(f(irc, msg, '$now'), '$now')
         n = f(irc, msg, '$randnick')
-        self.failUnless(n in irc.state.channels['#foo'].users)
+        self.assertTrue(n in irc.state.channels['#foo'].users)
         n = f(irc, msg, '$randomnick')
-        self.failUnless(n in irc.state.channels['#foo'].users)
+        self.assertTrue(n in irc.state.channels['#foo'].users)
         n = f(irc, msg, '$randomnick '*100)
         L = n.split()
-        self.failIf(all(L[0].__eq__, L), 'all $randomnicks were the same')
+        self.assertFalse(all(L[0].__eq__, L), 'all $randomnicks were the same')
         c = f(irc, msg, '$channel')
         self.assertEqual(c, msg.args[0])
 
